"""This package initializes the whole app with all of its modules.

This particular file additionally contains the applications factory.
"""

import os
import sys
from gitalizer.helpers.config import config
from gitalizer.extensions import logger, db, sentry, github

# Import models to load them into the declarative base of the current db engine
from gitalizer.models import * # noqa

# Entry point for click command line interaction
from gitalizer.cli import cli # noqa


class App:
    """A class representing the app."""

    def __init__(self, db, github, logger, sentry):
        """Create a new app."""
        self.db = db
        self.github = github
        self.logger = logger
        self.sentry = sentry


def create_app():
    """Create a new app."""
    # Check if the git clone dir can be created/accessed
    git_clone_dir = config['cloning']['temporary_clone_path']
    if not os.path.exists(git_clone_dir):
        try:
            os.makedirs(git_clone_dir)
        except PermissionError:
            logger.error(f"Gitalizer needs to have permissions to create the directory specified in 'GIT_CLONE_PATH': {git_clone_dir}")
            sys.exit(1)
    else:
        if not os.access(git_clone_dir, os.W_OK):
            logger.error(f"Gitalizer needs to have permissions to write to the directory specified in 'GIT_CLONE_PATH': {git_clone_dir}")
            sys.exit(1)

    return App(db, github, logger, sentry)


app = create_app()



"""Simple wrapper around github that allows for lazy initilization."""
from github import Github as ActualGithub


class Github(object):
    """Github wrapper class that allows single initialization."""

    def __init__(self, config):
        """Initialize github."""
        if config['github']['github_token']:
            user = config['github']['github_token']
            password = None
        else:
            user = config['github']['github_user']
            password = config['github']['github_password']
        self.github = ActualGithub(user, password)



"""Clean stuff from db, which occured through bugs."""
from datetime import datetime, timedelta
from sqlalchemy import or_, func
from sqlalchemy.orm import joinedload

from gitalizer.helpers.config import config
from gitalizer.extensions import logger
from gitalizer.models import (
    Repository,
    Commit,
    Contributor,
    Email,
)
from gitalizer.helpers.parallel import new_session
from gitalizer.helpers.parallel.manager import Manager
from gitalizer.helpers.parallel.list_manager import ListManager


def clean_db():
    """Clean stuff."""
    logger.info("Removing commits from fork repos.")
    session = new_session()
    try:
        all_repositories = session.query(Repository) \
            .filter(Repository.fork.is_(True)) \
            .filter(Repository.commits != None) \
            .options(joinedload(Repository.commits)) \
            .all()

        logger.info(f'Found {len(all_repositories)}')
        repositories_count = 0
        for repository in all_repositories:
            repository.commits = []
            session.add(repository)
            repositories_count += 1
            if repositories_count % 100 == 0:
                logger.info(f'Removed {repositories_count}')
                session.commit()

        logger.info("Remove unattached commits")
        session.query(Commit) \
            .filter(Commit.repositories == None) \
            .delete()
        session.commit()
    finally:
        session.close()


def complete_data():
    """Complete missing entities."""
    complete_repos()


def complete_repos():
    """Complete unfinished repsitories."""
    logger.info("Get unfinished or out of date repositories.")
    rescan_interval = int(config['aggregator']['repository_rescan_interval'])
    rescan_threshold = datetime.utcnow() - timedelta(seconds=rescan_interval)
    session = new_session()

    repos = session.query(Repository) \
        .filter(Repository.fork.is_(False)) \
        .filter(Repository.broken.is_(False)) \
        .filter(Repository.too_big.is_(False)) \
        .filter(or_(
            Repository.completely_scanned.is_(False),
            Repository.updated_at <= rescan_threshold,
        )) \
        .all()
    logger.info(f'Found {len(repos)}')

    full_names = [r.full_name for r in repos]
    repos_to_scan = set(full_names)

    manager = Manager('github_repository', repos_to_scan)
    manager.start()
    manager.run()

    session.close()


def update_data(update_all=False):
    """Update existing udata."""
    update_contributors(update_all)


def update_contributors(update_all):
    """Complete contributors."""
    session = new_session()
    logger.info(f'Start Scan.')

    # Look at the last two years
    time_span = datetime.now() - timedelta(days=2*365)

    results = session.query(Contributor, func.array_agg(Commit.sha)) \
        .filter(Contributor.login == Contributor.login) \
        .join(Email, Contributor.login == Email.contributor_login) \
        .join(Commit, or_(
            Commit.author_email_address == Email.email,
            Commit.committer_email_address == Email.email,
        )) \
        .filter(Commit.commit_time >= time_span) \
        .filter(or_(
            Contributor.location == None,
        )) \
        .group_by(Contributor.login) \
        .all()

    logger.info(f'Scanning {len(results)} contributors.')

    if update_all:
        contributors_to_scan = results
        logger.info(f'Scanning {len(contributors_to_scan)} contributors')
    else:
        count = 0
        contributors_to_scan = []
        for contributor, commits in results:
            if len(commits) > 100 and len(commits) < 20000:
                contributors_to_scan.append((contributor, commits))

            count += 1
            if count % 5000 == 0:
                logger.info(f'Found {count} contributors ({len(contributors_to_scan)} big)')

    manager = ListManager('github_user', contributors_to_scan)
    manager.start()
    manager.run()



"""Module containing the `User` model."""
from gitalizer.extensions import db


class TimezoneInterval(db.Model):
    """This class represents lol."""

    __tablename__ = 'timezone_interval'
    __table_args__ = (
        db.CheckConstraint('"start" < "end"'),
    )

    timezone = db.Column(db.String(120), nullable=False, primary_key=True)
    utcoffset = db.Column(db.Interval(), index=True)
    start = db.Column(db.DateTime(timezone=True), index=True, primary_key=True)
    end = db.Column(db.DateTime(timezone=True), index=True, primary_key=True)

    def __init__(self, timezone, offset, start, end):
        """Create a new TimezoneInterval."""
        self.timezone = timezone
        self.utcoffset = offset
        self.end = end
        self.start = start

    def __repr__(self):
        """Format a `TimezoneInterval` object."""
        return f'<TimezoneInterval {self.timezone}, {self.utcoffset}, {self.start}-{self.end}>'



"""Create a timezone database to reverse map timestamps to a timezone."""
import pendulum

from pendulum import timezones
from pendulum.tz.loader import Loader

from gitalizer.models import TimezoneInterval


def create_timezone_database(session):
    """Create new utc offset history for all timezones."""
    timezone_intervals = []
    for name in timezones:
        timezone = create_timezone(name)
        timezone_intervals += timezone

    for interval in timezone_intervals:
        new = TimezoneInterval(
            interval['name'],
            interval['utc-offset'],
            interval['start'],
            interval['end'],
        )
        session.add(new)

    session.commit()


def create_timezone(name: str):
    """Create a new timezone timeline."""
    test = Loader.load(name)

    transitions = test[0]

    intervals = []
    current_interval = None
    for transition in transitions:
        if not current_interval:
            current_interval = {
                'name': name,
                'utc-offset': transition.time - transition.utc_time,
                'start': transition.utc_time,
            }
            continue

        current_interval['end'] = transition.utc_time
        intervals.append(current_interval)

        current_interval = {
            'name': name,
            'utc-offset': transition.time - transition.utc_time,
            'start': transition.utc_time,
        }

    current_interval['end'] = pendulum.utcnow()

    return intervals



"""Helper for unified return messages between worker and manager."""
import traceback


def user_too_big_message(user_login: str):
    """Create an user_too_big_message."""
    return {
        'message': f'User {user_login} has too many repositories',
        'error': 'Too big',
    }


def user_up_to_date_message(user_login: str):
    """Create an user_up_to_date_message."""
    return {
        'message': f'User {user_login} is already up to date',
        'tasks': [],
    }


def error_message(message: str):
    """Return a error message with stacktrace."""
    return {
        'message': message,
        'error': traceback.format_exc(),
    }



"""Simple wrapper around sentry that allows for lazy initilization."""
from pygit2 import GitError
from raven import Client


class Sentry(object):
    """Sentry wrapper class that allows this app to work without a sentry token.

    If no token is specified in the config, the messages used for logging are simply not called.
    """

    initialized = False

    def __init__(self, config):
        """Construct new sentry wrapper."""
        if config['develop']['sentry_token'] is not None \
                and config['develop'].getboolean('sentry_enabled'):
            self.initialized = True
            self.sentry = Client(
                config['develop']['sentry_token'],
                ignore_exceptions=[
                    KeyboardInterrupt,
                    GitError,
                ],
            )

    def captureMessage(self, *args, **kwargs):
        """Capture message with sentry."""
        if self.initialized:
            self.sentry.captureMessage(*args, **kwargs)

    def captureException(self, *args, **kwargs):
        """Capture exception with sentry."""
        if self.initialized:
            self.sentry.captureException(*args, **kwargs)



"""Organization scanning cli."""
import sys
import click

from gitalizer.extensions import logger

from gitalizer.aggregator.github.organization import (
    get_github_organization,
    get_organization_memberships,
)


@click.group()
def organization():
    """Scan repositories."""
    pass


@click.command()
@click.argument('name')
def by_name(name):
    """Get github organizations for all known contributors."""
    try:
        logger.info(f'\nGet organization {name}')
        get_github_organization(name)
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


@click.command()
@click.argument('name')
def members(name):
    """Scan all members of a organization."""
    try:
        get_github_organization(name, True)
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


@click.command()
def user_membership():
    """Get all organizations for all known users (users in current database).

    This operation only scans the membership of users and doesn't scan any repositories.
    It is not yet multi processed.
    """
    try:
        get_organization_memberships()
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


organization.add_command(by_name)
organization.add_command(members)
organization.add_command(user_membership)



"""User scan."""
import sys
import click
from gitalizer.extensions import logger
from gitalizer.aggregator.github.user import (
    get_user,
    get_user_with_followers,
)


@click.command()
@click.argument('login')
@click.option('--with-followers', default=False, help='If true, not only the user but all followers and following will be scanned.')
def user(login, with_followers):
    """Get all repositories for a specific github user.

    If `--with-followers` is provided, the user and ALL of his followers/following will be scanned.
    This is recommended for a better coverage, but it will also take significantly longer and require significantly more disk space.
    """
    try:
        if with_followers:
            logger.info(f'\n\nGet friends of user {login}')
            get_user_with_followers(login)
        else:
            logger.info(f'\n\nGet user {login}')
            get_user(login)
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)



"""Module for multiprocessing management."""
import multiprocessing

from gitalizer.extensions import logger
from gitalizer.helpers.config import config
from gitalizer.helpers.parallel.task import Task
from gitalizer.helpers.parallel.worker import Worker


class Manager():
    """Class for managing various multiprocessing tasks."""

    def __init__(self, task_type: str, tasks: list,
                 sub_manager: 'Manager'=None):
        """Create a new manager."""
        self.tasks = set(tasks)
        self.task_type = task_type
        self.sub_manager = sub_manager
        self.started = False

        self.task_queue = multiprocessing.JoinableQueue()
        self.result_queue = multiprocessing.Queue()
        self.results = []
        if task_type == 'github_contributor':
            self.consumer_count = int(config['aggregator']['git_user_scan_threads'])
        elif task_type == 'github_repository':
            self.consumer_count = int(config['aggregator']['git_commit_scan_threads'])

    def start(self):
        """Initialize workers and add initial tasks."""
        # Create and start normal consumer
        consumers = [Worker(self.task_queue, self.result_queue)
                     for i in range(self.consumer_count)]
        for w in consumers:
            w.start()

        for task in self.tasks:
            self.task_queue.put(Task(self.task_type, task))
        self.started = True

    def add_tasks(self, tasks: list):
        """Add some tasks to the queue."""
        # Add unique tasks to queue
        tasks = set(tasks)
        if self.started:
            for task in (tasks - self.tasks):
                self.task_queue.put(Task(self.task_type, task))

        # Add new tasks to task set.
        self.tasks |= tasks

    def run(self):
        """All tasks are added. Process worker responses and wait for worker to finish."""
        # Start the sub manager
        if self.sub_manager is not None:
            logger.info('Start sub manager.')

        # Poison pill for user scanner
        logger.info('Add poison pills.')
        for _ in range(self.consumer_count+1):
            self.task_queue.put(None)

        logger.info(f'Processing {len(self.tasks)} tasks')
        finished_tasks = 0
        while finished_tasks < len(self.tasks):
            logger.info(f'Waiting: {finished_tasks} of {len(self.tasks)}')
            result = self.result_queue.get()
            self.results.append(result)

            logger.info(result['message'])
            if 'error' in result:
                logger.info('Encountered an error:')
                logger.info(result['error'])
            elif self.sub_manager is not None:
                self.sub_manager.add_tasks(result['tasks'])
            finished_tasks += 1

        # All sub tasks have been added.
        # Wait for them to finish.
        if self.sub_manager is not None:
            self.sub_manager.start()
            self.sub_manager.run()



"""Representation of analysis results."""
from uuid import uuid4
from sqlalchemy import ForeignKey
from sqlalchemy.dialects.postgresql import UUID, JSONB

from gitalizer.extensions import db


class AnalysisResult(db.Model):
    """Analysis results."""

    __tablename__ = 'analysis_result'

    id = db.Column(UUID(as_uuid=True), primary_key=True, default=uuid4)
    contributer_login = db.Column(
        db.String(240),
        ForeignKey('contributor.login', ondelete='CASCADE',
                   onupdate='CASCADE', deferrable=True),
        index=True)

    # Timezone check
    commit_count = db.Column(db.Integer())
    timezone_switches = db.Column(db.Integer())
    different_timezones = db.Column(db.Integer())
    last_change = db.Column(db.DateTime())
    intermediate_results = db.Column(JSONB)

    # Relationships
    contributor = db.relationship("Contributor", back_populates="analysis_result")



"""Representation of a git repository."""

from datetime import datetime, timedelta
from sqlalchemy import ForeignKey, func
from sqlalchemy.orm import backref
from sqlalchemy.orm.collections import attribute_mapped_collection

from gitalizer.helpers.config import config
from gitalizer.extensions import db
from gitalizer.models.commit import commit_repository
from gitalizer.models.contributor import contributor_repository


class Repository(db.Model):
    """Repository model."""

    __tablename__ = 'repository'

    clone_url = db.Column(db.String(240), primary_key=True)
    parent_url = db.Column(
        db.String(240),
        ForeignKey('repository.clone_url', ondelete='SET NULL',
                   onupdate='CASCADE', deferrable=True),
        index=True,
    )
    name = db.Column(db.String(240))
    full_name = db.Column(db.String(240), unique=True)
    created_at = db.Column(db.DateTime(timezone=True))

    fork = db.Column(db.Boolean(), default=False,
                     server_default='FALSE', nullable=False)
    broken = db.Column(db.Boolean(), default=False,
                       server_default='FALSE', nullable=False)
    too_big = db.Column(db.Boolean(), default=False,
                        server_default='FALSE', nullable=False)
    completely_scanned = db.Column(db.Boolean(), default=False,
                                   server_default='FALSE', nullable=False)
    updated_at = db.Column(db.DateTime, server_default=func.now(), nullable=False)

    children = db.relationship(
        "Repository",
        backref=backref('parent', remote_side=[clone_url]),
    )

    commits = db.relationship(
        "Commit",
        secondary=commit_repository,
        back_populates="repositories",
    )
    commits_by_hash = db.relationship(
        "Commit",
        collection_class=attribute_mapped_collection('sha'),
        secondary=commit_repository,
    )

    contributors = db.relationship(
        "Contributor",
        secondary=contributor_repository,
        back_populates="repositories",
    )

    def __init__(self, clone_url, name=None, full_name=None):
        """Constructor."""
        self.clone_url = clone_url
        self.name = name
        self.full_name = full_name

    @staticmethod
    def get_or_create(session, clone_url: str, name=None, full_name=None):
        """Get an existing repository from db or create a new one."""
        repo = session.query(Repository).get(clone_url)

        if not repo:
            repo = Repository(clone_url, name, full_name)
            session.add(repo)
            session.commit()

        return repo

    def should_scan(self):
        """Check if the repo has been updated in the last day.

        If that is the case, we want to skip it.
        """
        rescan_interval = int(config['aggregator']['repository_rescan_interval'])
        rescan_threshold = datetime.utcnow() - timedelta(seconds=rescan_interval)
        up_to_date = self.completely_scanned and self.updated_at >= rescan_threshold

        if self.fork or self.broken or self.too_big or up_to_date:
            return False

        return True

    def is_invalid(self):
        """Check if we should skip this repository for for checking."""
        return (self.broken or self.too_big)



"""Helper for time zone handling."""
from pytz import country_timezones
from geopy.geocoders import Nominatim


timezone_country = {}
for countrycode in country_timezones:
    timezones = country_timezones[countrycode]
    for timezone in timezones:
        timezone_country[timezone] = countrycode


def map_timezone_to_state(name):
    """Map a timezone to a state."""
    components = name.split('/')

    # If there are three components, the middle component is the state
    if len(components) == 3:
        return components[1]

    try:
        search_string = components[1].replace('_', ' ')
        geolocator = Nominatim()
        location = geolocator.geocode(search_string)
        location = geolocator.reverse([location.raw['lat'], location.raw['lon']])

        state = location.raw['address']['state']
        return state

    except BaseException:
        return None


def map_timezone_to_utc(name):
    """Convert a timezone like GMT or MET to an utc offset timezone name."""
    # Check if we have GMT
    if 'GMT' in name:
        # If we have GMT, get the offset, invert it and create proper utc name
        offset = name.split('GMT')[1]
        if '-' in offset:
            offset.replace('-', '+')
        elif '+' in offset:
            offset.replace('+', '-')
        else:
            offset = 0
        return offset

    # No compatible zone found. Go back to default name
    else:
        return name



"""Repository scanning cli."""
import sys
import click

from gitalizer.extensions import logger
from gitalizer.aggregator.github.repository import (
    get_github_repository_by_owner_name,
    get_github_repository_users,
)


@click.group()
def repository():
    """Scan repositories."""
    pass


@click.command()
@click.argument('owner')
@click.argument('repository')
def from_github(owner, repository):
    """Get a github repository by owner and name."""
    try:
        logger.info(f'\n\nGet {repository} from user {owner}')
        get_github_repository_by_owner_name(owner, repository)
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


@click.command()
@click.argument('full_name')
def from_github_for_users(full_name):
    """Scan all users of a github repository."""
    try:
        logger.info(f'\nGet all users from {full_name}')
        get_github_repository_users(full_name)
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


repository.add_command(from_github)
repository.add_command(from_github_for_users)



"""Representation of a git commit."""
from sqlalchemy import ForeignKey, CheckConstraint

from gitalizer.extensions import db


commit_repository = db.Table(
    'commit_repository',
    db.Column('commit_sha',
              db.String(40),
              ForeignKey('commit.sha', ondelete='CASCADE',
                         onupdate='CASCADE', deferrable=True),
              index=True),
    db.Column('repository_clone_url',
              db.String(240),
              ForeignKey('repository.clone_url', ondelete='CASCADE',
                         onupdate='CASCADE', deferrable=True),
              index=True),
    db.UniqueConstraint('repository_clone_url', 'commit_sha'),
)


class Commit(db.Model):
    """Commit model."""

    __tablename__ = 'commit'
    __table_args__ = (
        CheckConstraint(
            "(additions is NULL and deletions is NULL) or "
            "(additions is not NULL and deletions is not NULL)",
        ),
    )

    sha = db.Column(db.String(40), primary_key=True)
    commit_time = db.Column(db.DateTime(timezone=True))
    commit_time_offset = db.Column(db.Interval())
    creation_time = db.Column(db.DateTime(timezone=True))
    creation_time_offset = db.Column(db.Interval())
    additions = db.Column(db.Integer())
    deletions = db.Column(db.Integer())

    # Email addresses
    author_email_address = db.Column(
        db.String(240), ForeignKey('email.email'),
        index=True, nullable=False,
    )
    committer_email_address = db.Column(
        db.String(240), ForeignKey('email.email'),
        index=True, nullable=False,
    )
    author_email = db.relationship(
        "Email", back_populates="author_commits",
        foreign_keys=[author_email_address],
    )
    committer_email = db.relationship(
        "Email", back_populates="committer_commits",
        foreign_keys=[committer_email_address],
    )

    repositories = db.relationship(
        "Repository",
        secondary=commit_repository,
        back_populates="commits",
    )

    def __init__(self, sha, repository, author_email, committer_email):
        """Constructor."""
        self.sha = sha
        self.author_email = author_email
        self.committer_email = committer_email
        self.repositories.append(repository)

    def local_time(self):
        """Get the local commit time for this commit."""
        new_time = self.commit_time
        if self.commit_time.utcoffset():
            new_time = self.commit_time - self.commit_time.utcoffset()
            new_time = new_time.replace(tzinfo=None)
            new_time += self.commit_time_offset

        return new_time



"""Representation of a git repository contributor."""

from datetime import datetime, timedelta, timezone
from sqlalchemy import ForeignKey
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import joinedload

from gitalizer.helpers.config import config
from gitalizer.extensions import db, logger


contributor_repository = db.Table(
    'contributor_repository',
    db.Column('contributor_login',
              db.String(240),
              ForeignKey('contributor.login', ondelete='CASCADE',
                         onupdate='CASCADE', deferrable=True),
              index=True),
    db.Column('repository_clone_url',
              db.String(240),
              ForeignKey('repository.clone_url', ondelete='CASCADE',
                         onupdate='CASCADE', deferrable=True),
              index=True),
    db.UniqueConstraint('repository_clone_url', 'contributor_login'),
)


contributor_organizations = db.Table(
    'contributor_organizations',
    db.Column('contributor_login',
              db.String(240),
              ForeignKey('contributor.login', ondelete='CASCADE',
                         onupdate='CASCADE', deferrable=True),
              index=True),
    db.Column('organization_login',
              db.String(240),
              ForeignKey('organization.login', ondelete='CASCADE',
                         onupdate='CASCADE', deferrable=True),
              index=True),
    db.UniqueConstraint('contributor_login', 'organization_login'),
)


class Contributor(db.Model):
    """Contributor model."""

    __tablename__ = 'contributor'

    login = db.Column(db.String(240), primary_key=True, nullable=False)
    emails = db.relationship("Email", back_populates="contributor")
    location = db.Column(db.String(240))

    # Relationships
    repositories = db.relationship(
        "Repository",
        secondary=contributor_repository,
        back_populates="contributors")
    organizations = db.relationship(
        "Organization",
        secondary=contributor_organizations,
        back_populates="contributors")
    analysis_result = db.relationship(
        "AnalysisResult",
        uselist=False,
        back_populates="contributor")

    too_big = db.Column(db.Boolean, default=False, server_default='FALSE', nullable=False)
    last_full_scan = db.Column(db.DateTime(timezone=True))

    def __init__(self, login: str):
        """Constructor."""
        self.login = login

    def __repr__(self):
        """Format a `Contributor` object."""
        return f'<Contributor {self.login}>'

    @staticmethod
    def get_contributor(login: str, session, eager_repositories=False, do_commit=True):
        """Create new contributor or add repository to it's list.

        Try multiple times, as we can get Multiple additions through threading.
        """
        _try = 0
        tries = 3
        exception = None
        while _try <= tries:
            try:
                contributor = session.query(Contributor)
                if eager_repositories:
                    contributor.options(joinedload(Contributor.repositories))
                contributor = contributor.get(login)
                if not contributor:
                    # Commit to prevent data loss in case we get an
                    # integrity error and need to rollback.
                        contributor = Contributor(login)
                session.add(contributor)
                if do_commit:
                    session.commit()
                return contributor
            except IntegrityError as e:
                logger.error(f'Got an Contributor IntegrityError, Try {_try} of {tries}')
                session.rollback()
                _try += 1
                exception = e
                pass

        raise exception

    def should_scan(self):
        """Check if the user has been scanned in the last day.

        If that is the case, we want to skip it.
        """
        no_repositories = len(self.repositories) == 0
        if no_repositories or self.last_full_scan is None:
            return True

        rescan_interval = int(config['aggregator']['contributor_rescan_interval'])
        rescan_threshold = datetime.utcnow(timezone.utc) - timedelta(seconds=rescan_interval)
        up_to_date = self.last_full_scan and self.last_full_scan >= rescan_threshold

        if not up_to_date:
            return True

        return False



"""DB cli interface."""
import click
from sqlalchemy_utils.functions import (
    database_exists,
    create_database,
    drop_database,
)

from gitalizer.models import (
    TimezoneInterval,
)
from gitalizer.extensions import db as real_db
from gitalizer.timezone.database import create_timezone_database


@click.group()
def db():
    """DB subcommand base."""
    pass


@db.command()
def initdb():
    """Initialize the database.

    This will drop and recreate the actual database if it already exists.
    The database name from the config is used for this.
    """
    # If there is an existing DB, make sure to drop it and start completely fresh.
    db_url = real_db.engine.url
    if database_exists(db_url):
        drop_database(db_url)
    create_database(db_url)

    real_db.Model.metadata.create_all()
    session = real_db.session
    create_timezone_database(session)
    session.commit()


@db.command()
def build_time_db():
    """Drop and recreate all utc offset timeinterval database entries."""
    real_db.create_all()

    session = real_db.get_session()
    session.query(TimezoneInterval).delete()
    create_timezone_database(session)
    session.commit()
    session.close()



"""User scan."""
import click

from gitalizer.extensions import db
from gitalizer.models import (
    Commit,
    commit_repository,
    Repository,
)


@click.group()
def delete():
    """Cli wrapper for the database delete command group."""
    pass


@click.command()
@click.argument('full_name')
def repository(full_name):
    """Delete a specific repository."""
    session = db.get_session()
    try:
        repository = session.query(Repository) \
            .filter(Repository.full_name == full_name) \
            .one()

        commit_shas = session.query(Commit.sha) \
            .join(
                commit_repository,
                commit_repository.c.repository_clone_url == repository.clone_url,
            ) \
            .filter(commit_repository.c.commit_sha == Commit.sha) \
            .all()

        commit_shas = [c[0] for c in commit_shas]
        if commit_shas:
            session.query(Commit) \
                .filter(Commit.sha.in_(commit_shas)) \
                .delete(synchronize_session=False)

        session.query(Repository) \
            .filter(Repository.full_name == full_name) \
            .delete()
        session.commit()
    finally:
        session.close()


delete.add_command(repository)



"""Queue processing workers."""
import time
import multiprocessing

from gitalizer.extensions import sentry


class Worker(multiprocessing.Process):
    """A worker which gets tasks from a queue."""

    def __init__(self, task_queue, result_queue):
        """Create a new worker."""
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self):
        """Process incoming tasks."""
        while True:
            try:
                next_task = self.task_queue.get()
                # Poison pill received: Shutdown
                if next_task is None:
                    self.task_queue.task_done()
                    break
                answer = next_task()
                self.task_queue.task_done()
                self.result_queue.put(answer)
                time.sleep(1)
            except KeyboardInterrupt:
                break
            except BaseException:
                sentry.captureException()
        return



"""Representation of a git author email."""

from sqlalchemy import ForeignKey
from sqlalchemy.exc import IntegrityError

from gitalizer.extensions import db


class Email(db.Model):
    """Email model."""

    __tablename__ = 'email'

    email = db.Column(db.String(240), primary_key=True)

    # Contributor
    contributor_login = db.Column(
        db.String(240), ForeignKey('contributor.login'), index=True,
    )
    contributor = db.relationship("Contributor", back_populates="emails")

    # Commits relationships
    author_commits = db.relationship(
        "Commit", back_populates="author_email",
        foreign_keys='Commit.author_email_address',
    )
    committer_commits = db.relationship(
        "Commit", back_populates="committer_email",
        foreign_keys='Commit.committer_email_address',
    )
    unknown = db.Column(db.Boolean(), default=False)

    def __init__(self, email, contributor=None):
        """Constructor."""
        self.email = email
        self.contributor = contributor

    @staticmethod
    def get_email(email_address: str, session, do_commit=True):
        """Create new email.

        Try multiple times, as we can get Multiple additions through threading.
        """
        _try = 0
        tries = 3
        exception = None
        while _try <= tries:
            try:
                email = session.query(Email).get(email_address)
                if not email:
                    email = Email(email_address)
                    session.add(email)
                    if do_commit:
                        session.commit()
                return email
            except IntegrityError as e:
                print(f'Got an Email IntegrityError, Try {_try} of {tries}')
                session.rollback()
                _try += 1
                exception = e
                pass

        raise exception



"""Helper functions to manage sessions between processes."""

from sqlalchemy.orm import sessionmaker
from gitalizer.extensions import db


def new_session():
    """Create a new session."""
    session = sessionmaker(bind=db.engine)()
    return session


def create_chunks(l, n):
    """Chunk a list into n sized chunks."""
    n = max(1, n)
    return [l[i:i+n] for i in range(0, len(l), n)]



"""Cli for gitalizer."""
import click
from .db import db
from .delete import delete
from .maintenance import maintenance
from .organization import organization
from .user import user
from .repository import repository


@click.group()
def cli():
    """Init function for gitalizer."""
    pass


@click.group()
def scan():
    """Init function for gitalizer."""
    pass


cli.add_command(db)
cli.add_command(scan)
cli.add_command(delete)
cli.add_command(maintenance)

scan.add_command(user)
scan.add_command(repository)
scan.add_command(organization)



"""A task for the multiprocessing worker."""


class Task(object):
    """An object which is passed to a worker."""

    def __init__(self, task_type, task):
        """Create a new object."""
        self.task = task
        self.task_type = task_type

    def __call__(self):
        """Actual work logic."""
        if self.task_type == 'github_contributor':
            from gitalizer.aggregator.github.user import get_user_repos
            return get_user_repos(self.task)
        elif self.task_type == 'github_repository':
            from gitalizer.aggregator.github.repository import get_github_repository
            return get_github_repository(self.task)
        elif self.task_type == 'github_user':
            from gitalizer.aggregator.github.user import get_user_data
            return get_user_data(self.task)



"""Setuptools file for gitalizer."""
from setuptools import setup, find_packages

setup(
    name='gitalizer',
    author='Arne Beer',
    author_email='privat@arne.beer',
    version='0.1.3',
    description='A git/github repository data collector for Git to analyse privacy implications of exposing Git metadata',
    keywords='git aggregator metadata github',
    url='http://github.com/nukesor/gitalizer',
    license='MIT',
    install_requires=[
        # Base
        'alembic~=0.8',
        'SQLAlchemy~=1.1.4',
        'sqlalchemy-utils~=0.32',
        'psycopg2-binary',

        # Aggregator
        'pygithub~=1.35',
        'pygit2~=0.27',
        'gitpython~=2.1',
        'pytz~=2018.3',

        # Logging
        'raven',

        # Time handling
        'pendulum~=1.4',
        'pytzdata~=2018.3',

        # Location handling
        'geopy',

        # Cli
        'click',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Environment :: Console',
    ],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'gitalizer=gitalizer:cli',
        ],
    })



from .commit import Commit, commit_repository  # noqa
from .email import Email  # noqa
from .organization import Organization # noqa
from .repository import Repository  # noqa
from .contributor import Contributor, contributor_repository  # noqa
from .time import TimezoneInterval # noqa
from .analysis_result import AnalysisResult # noqa



"""Logger class."""
import os
import sys
import logging

from logging.handlers import RotatingFileHandler


def init_logging(config):
    """Create log directory and initialize logger."""
    # Create log directory, if it doesn't exist
    logger = logging.getLogger('gitalizer')
    log_dir = config['develop']['log_dir']
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    logger.setLevel(logging.INFO)

    # Initialize log file for daemon output
    daemon_log_path = os.path.join(log_dir, 'gitalizer.log')

    log_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    log_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    # Add a handler which outputs the log to stdout
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(log_format)
    logger.addHandler(stdout_handler)

    # Add a handler which outputs to a logfile
    file_handler = RotatingFileHandler(
        daemon_log_path,
        maxBytes=(1048576*5),
        backupCount=7,
    )
    file_handler.setFormatter(log_format)
    logger.addHandler(file_handler)

    return logger



"""Sqlalchemy helper."""
import functools
import sqlalchemy
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Query
from sqlalchemy.ext.declarative import declarative_base


class DB(object):
    """Create an engine for easy session creation."""

    def __init__(self, config):
        """Lazy initializer which takes an `app` and sets up sentry."""
        self.engine = create_engine(
            config['database']['uri'],
            echo=config['database'].getboolean('echo'),
        )
        self.Model = declarative_base(bind=self.engine)

        query_class = Query
        # Inject sqlalchemy into db helper class
        for module in sqlalchemy, sqlalchemy.orm:
            for key in module.__all__:
                if not hasattr(self, key):
                    setattr(self, key, getattr(module, key))
        # Note: self.Table does not attempt to be a SQLAlchemy Table class.
        self.Table = _make_table(self)
        self.relationship = _wrap_with_default_query_class(self.relationship, query_class)
        self.relation = _wrap_with_default_query_class(self.relation, query_class)
        self.dynamic_loader = _wrap_with_default_query_class(self.dynamic_loader, query_class)

    @property
    def session(self):
        """Short handle for getting a session.

        Also needed for backwards compatibility during refactoring.
        """
        return self.get_session()

    def get_session(self):
        """Create a new session."""
        session = sessionmaker(bind=self.engine)()
        return session

#    def Table(self, *args, **kwargs):
#        """Creating a sqlalchemy table with the current db metadata."""
#        if len(args) > 1 and isinstance(args[1], self.Column):
#            args = (args[0], self.Model.metadata) + args[1:]
#        info = kwargs.pop('info', None) or {}
#        info.setdefault('bind_key', None)
#        kwargs['info'] = info
#        return sqlalchemy.Table(*args, **kwargs)


def _set_default_query_class(d, query_class):
    if 'query_class' not in d:
        d['query_class'] = query_class


def _wrap_with_default_query_class(fn, query_class):
    @functools.wraps(fn)
    def newfn(*args, **kwargs):
        _set_default_query_class(kwargs, query_class)
        if "backref" in kwargs:
            backref = kwargs['backref']
            if isinstance(backref, str):
                backref = (backref, {})
            _set_default_query_class(backref[1], query_class)
        return fn(*args, **kwargs)
    return newfn


def _make_table(db):
    def _make_table(*args, **kwargs):
        if len(args) > 1 and isinstance(args[1], db.Column):
            args = (args[0], db.Model.metadata) + args[1:]
        info = kwargs.pop('info', None) or {}
        info.setdefault('bind_key', None)
        kwargs['info'] = info
        return sqlalchemy.Table(*args, **kwargs)
    return _make_table



"""This module contains various extensions which will serve as a globally initialized import location."""
from gitalizer.helpers.config import config
from .github import Github
from .sentry import Sentry
from .logger import init_logging
from .sqlalchemy import DB

db = DB(config)
github = Github(config)
sentry = Sentry(config)
logger = init_logging(config)



"""Module containing the `Organization` model."""
from sqlalchemy.exc import IntegrityError

from gitalizer.extensions import db
from gitalizer.models.contributor import contributor_organizations


class Organization(db.Model):
    """Organization model."""

    __tablename__ = 'organization'

    login = db.Column(db.String(240), primary_key=True)
    url = db.Column(db.String(240))

    contributors = db.relationship(
        "Contributor",
        secondary=contributor_organizations,
        back_populates="organizations")

    def __init__(self, login, url):
        """Construct an `Organization`."""
        self.login = login
        self.url = url

    @staticmethod
    def get_organization(login: str, url: str, session):
        """Create new organization or add repository to it's list.

        Try multiple times, as we can get Multiple additions through threading.
        """
        _try = 0
        tries = 3
        exception = None
        while _try <= tries:
            try:
                organization = session.query(Organization).get(login)
                if not organization:
                    organization = Organization(login, url)
                session.add(organization)
                session.commit()
                return organization
            except IntegrityError as e:
                print(f'Got an Organization IntegrityError, Try {_try} of {tries}')
                _try += 1
                exception = e
                pass

        raise exception



"""DB maintenance operations."""

import sys
import click

from gitalizer.extensions import logger
from gitalizer.helpers.db.maintenance import (
    clean_db,
    complete_data,
    update_data,
)


@click.group()
def maintenance():
    """Cli wrapper for the maintenance command group."""
    pass


@click.command()
def clean():
    """Remove uncomplete or unwanted data.

    This includes:
        - Duplicate commits from forked repositories (Shouldn't happen any longer).
        - Dangling commits (Shouldn't happen any longer).
    """
    try:
        clean_db()
    except KeyboardInterrupt:
        sys.exit(1)


@click.command()
def complete():
    """Complete missing data from previous runs.

    This includes:
        - Complete all unfinished repositories
    """
    try:
        complete_data()
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


@click.command()
@click.option('--update-all', default=False, help='Scan EVERYTHING again.')
def update(update_all):
    """Update data from previous runs.

    If the `--update-all` flag is provided, everything will be scanned again.
    For instance, without the flag, only contributors with more than 100 commits will be scanned.
    This prevents scans of users which might be uninteresting peers and, for instance, only have one commit in a uninteresting repository.

    This includes:
        - Update all contributors
    """
    try:
        update_data(update_all)
    except KeyboardInterrupt:
        logger.info("CTRL-C Exiting Gracefully")
        sys.exit(1)


maintenance.add_command(clean)
maintenance.add_command(complete)
maintenance.add_command(update)
